//----------------------------------------------------------------------------
/// @file benchmark_sort_05.cpp
/// @brief Benchmark of several sort methods with different sets of data, and
///        data of different sizes
///
/// @author Copyright (c) 2010 2015 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//-----------------------------------------------------------------------------
#include <parallel/basic_iterator.h>
#include <parallel/features.h>
#include <parallel/parallel.h>
#include <parallel/algorithm>
#include <omp.h>
#include <stdlib.h>
#include <iostream>
#include <random>
#include <sys/time.h>
#include <sys/resource.h>
#include <vector>
#include "tbb/tbb_stddef.h"
#include "tbb/task_scheduler_init.h"
#include "tbb/parallel_sort.h"


#include <boost/countertree/util_sort.hpp>
#include <boost/countertree/util/time_measure.hpp>


#define NELEM 100000000

using namespace std ;
using namespace __gnu_parallel;
using namespace boost::countertree::util::sort ;
using boost::countertree::NThread ;
using boost::countertree::util::time_point ;
using boost::countertree::util::now;
using boost::countertree::util::subtract_time ;

std::mt19937_64 my_rand(0);

template <class RBRC> void Generator  ( uint64_t N);
template <class RBRC> void Prueba  ( vector<RBRC> &V2);

template <uint32_t NN>
struct reburcio
{   uint64_t M[NN];

    reburcio ( const reburcio &R ) { M[0] = R.M[0] ; };
    reburcio ( uint64_t K =0 ) { M[0] = K ; };
    //reburcio & operator = ( const reburcio &R)
    //{   M[0] = R.M[0];
    //    return *this ;
    //};
    reburcio & operator = ( uint64_t K)
    {   M[0] = K;
        return *this ;
    };
    bool operator < ( const reburcio &R) const{ return (M[0]<R.M[0]);};
};


int main ()
{   //---------------------- begin --------------------------------
    cout<<"max. OpenMP threads:"<<omp_get_max_threads();
    cout<<" (processors)"<< omp_get_num_procs()<<endl;
    Generator< reburcio<1> > ( NELEM);
    Generator< reburcio<2> > ( NELEM>>1);
    Generator< reburcio<4> > ( NELEM>>2);
    Generator< reburcio<6> > ( NELEM /6);
    Generator< reburcio<8> > ( NELEM>>3);
    Generator< reburcio<16> > ( NELEM>>4);
    Generator< reburcio<32> > ( NELEM>>5);
}

template <class RBRC>
void Generator (uint64_t N )
{   //--------------------------- Inicio --------------------------
    vector<RBRC> A ;
   	A.reserve ( N);

    //------------------------ Inicio -----------------------------
    cout<<"\n***************************************************************\n";
    cout<<"Sort of N = "<<N<<" elements of size "<<sizeof (RBRC)<<"\n";
    cout<<"*****************************************************************\n";
    cout<<"Sorted elements -----------------------------------------------\n";
    A.clear() ;
    for ( uint32_t i = 0 ; i < N ; i ++)
        A.emplace_back( N +i) ;
    Prueba( A) ;

    cout<<"Reverse sorted elements ---------------------------------------\n";
    A.clear() ;
    for ( uint32_t i = 0 ; i < N ; i ++)
        A.emplace_back( (N<<1) -i) ;
    Prueba(A) ;

    cout<<"Random elements, few repeated ( rand() )-----------------------\n";
    A.clear();
    for ( uint32_t i = 0 ; i < N  ; i ++)
        A.emplace_back(my_rand()) ;
    Prueba(A) ;

    cout<<"Random elements, quite repeated ( rand()  % (N/2) )------------\n";
    A.clear() ;
    for ( uint32_t i = 0 ; i < N  ; i ++)
        A.emplace_back(my_rand()  % (N/2));
    Prueba(A) ;

    cout<<"Random element many repeated (rand() % 10000)------------------\n";
    A.clear() ;
    for ( uint32_t i = 0 ; i < N  ; i ++)
        A.emplace_back(my_rand() % 10000);
    Prueba(A) ;

    cout<<"Equal elements ------------------------------------------------\n";
    A.clear() ;
    for ( uint32_t i = 0 ; i < N  ; i ++) A.emplace_back( N);
    Prueba(A) ;

};


template <class RBRC>
void Prueba  ( std::vector <RBRC> & A)
{   //--------------------------- Inicio --------------------------
    double duracion ;
	time_point start, finish;
	NThread  NT ;


    std::vector <RBRC> B ;
    B.reserve ( A.size());

	//cout<<"--------------------- STL std::sort -------------------------\n";
	B = A ;
	start= now() ;
	std::sort (B.begin() , B.end()  );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    cout<<"STL std::sort :"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (B.size() -1) ; i ++ )
	{	if ( B[i+1] < B[i] ) cout<<"Error Pos :"<<i ;
	};


	//cout<<"--------------------- Parallel std::sort ---------------------\n";
	B = A ;
	start= now() ;
	__gnu_parallel::sort (B.begin() , B.end()  );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
	cout<<"parallel::sort :"<<duracion<<" secs "<<std::endl;
	for ( unsigned i = 0 ; i < (B.size() -1) ; i ++ )
	{	if ( B[i+1] < B[i] ) cout<<"Error Pos :"<<i ;
	};


	//cout<<"---------------------tbb::parallel_sort ----------------------\n";
	B = A ;
	start= now() ;
	tbb::parallel_sort (B.begin() , B.end()  );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    cout<<"tbb::parallel_sort :"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (B.size() -1) ; i ++ )
	{	if ( B[i+1] < B[i] ) cout<<"Error Pos :"<<i ;
	};

	//cout<<"------- boost::countertree::parallel_sort --- ----------------\n";
	B = A ;
	start= now() ;
	boost::countertree::parallel_sort (B.begin() , B.end() ,NThread() );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    cout<<"countertree::parallel_sort:"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (B.size() -1) ; i ++ )
	{	if ( B[i+1] < B[i] ) cout<<"Error Pos :"<<i ;
	};

	//cout<<"----------- STL std::stable_sort ----------------------------\n";
	cout<<"\n";
	B = A ;
	start= now() ;
	std::stable_sort (B.begin() , B.end()  );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    cout<<"std::stable_sort :"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (B.size() -1) ; i ++ )
	{	if ( B[i+1] < B[i] ) cout<<"Error Pos :"<<i ;
	};



	//cout<<"------------------- Parallel std::stable_sort -----------------------\n";
	B = A ;
	start= now() ;
	__gnu_parallel::stable_sort (B.begin() , B.end()  );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
	cout<<"parallel::stable_sort :"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (B.size() -1) ; i ++ )
	{	if ( B[i+1] < B[i] ) cout<<"Error Pos :"<<i ;
	};


	//cout<<"----------------- parallel_merge_sort ------------------------\n";
	B = A ;
	start= now() ;
	boost::countertree::parallel_merge_sort (B.begin() , B.end() ,NThread() );
    finish = now() ;
	duracion = subtract_time(finish ,start) ;
    cout<<"countertree::parallel_merge_sort :"<<duracion<<" secs\n";
	for ( unsigned i = 0 ; i < (B.size() -1) ; i ++ )
	{	if ( B[i+1] < B[i] ) cout<<"Error Pos :"<<i ;
	};
    cout<<endl;

};
