//----------------------------------------------------------------------------
/// @file atomic.hpp
/// @brief Basic functions definitions and data structures for the parallel process
/// @author Copyright (c) 2010 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//-----------------------------------------------------------------------------
#ifndef __CNTREE_UTIL_ATOMIC_HPP
#define __CNTREE_UTIL_ATOMIC_HPP

#include <thread>
#include <atomic>
#include <cassert>
#include <utility>
#include <type_traits>
#include <boost/countertree/util/config.hpp>
#include <boost/countertree/util/definition.hpp>
#include <boost/countertree/util/algorithm.hpp>
#include <boost/countertree/util/util_iterator.hpp>


namespace boost
{
namespace countertree
{
namespace util
{

//---------------------------------------------------------------------------
//         D E F I N I T I O N S
//---------------------------------------------------------------------------

//---------------------------------------------------------------------------
//   A T O M I C  I N C R E M E N T    A N D   R E A D
//---------------------------------------------------------------------------
template <typename T >
inline T atomic_read ( std::atomic<T> & C)
{   return std::atomic_load_explicit (&C, std::memory_order_acquire);
};
template <typename T, typename T2>
inline T atomic_add ( std::atomic<T> &C, T2 N1)
{   static_assert ( std::is_integral <T2>::value , "Bad parameter");
    return std::atomic_fetch_add_explicit (&C,(T)N1,std::memory_order_acq_rel);
};
template <typename T, typename T2>
inline T atomic_sub ( std::atomic<T> &C, T2 N1)
{   static_assert ( std::is_integral <T2>::value , "Bad parameter");
    return std::atomic_fetch_sub_explicit (&C,(T)N1,std::memory_order_acq_rel);
};

template <typename T, typename T2>
inline void atomic_write ( std::atomic<T> &C, T2 N1)
{   static_assert ( std::is_integral <T2>::value , "Bad parameter");
    return std::atomic_store_explicit (&C, (T)N1, std::memory_order_release);
};


static const uint32_t NThread_HW = std::thread::hardware_concurrency();
class NThread
{   uint32_t NT ;

public:

    NThread ( void): NT ( NThread_HW){};
    NThread ( const NThread & Number):NT ( Number.NT){};
    NThread ( uint32_t Nmb):NT ((Nmb< 2)?1:Nmb ){};

    uint32_t operator () ( void)const { return NT;};

    NThread & operator = (const NThread & Number)
    {   NT = Number.NT ;
        return *this ;
    };
    NThread & operator= ( uint32_t Nmb)
    {   NT = ( Nmb < 2)?1:Nmb;
        return *this ;
    };
};

//****************************************************************************
}; // End namespace util
}; // End namespace countertree
}; // Fin namespace boost
//****************************************************************************
#endif
