//----------------------------------------------------------------------------
/// @file   recursive_mutex_1WNR.hpp
/// @brief  This file contains the implementation of the several types of
///         recursive fastmutex for read and write
///
/// @author Copyright (c) 2010 2013 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//-----------------------------------------------------------------------------
#ifndef __CNTREE_UTIL_MUTEX_HPP
#define __CNTREE_UTIL_MUTEX_HPP

#include <cntree/util/spinlock.hpp>
#include <mutex>
#include <cassert>
#include <iostream>

namespace boost
{
namespace countertree
{
namespace util
{

//
//###########################################################################
//                                                                         ##
//    ################################################################     ##
//    #                                                              #     ##
//    #                      C L A S S                               #     ##
//    #             E M P T Y _ M U T E X _ D A T A                  #     ##
//    #                                                              #     ##
//    ################################################################     ##
//                                                                         ##
//###########################################################################
//
//---------------------------------------------------------------------------
/// @class  empty_mutex_data
/// @brief  This class implements information for to be used by
///         fast_read_cnc and fastmutex_write_cnc in order to block the code\n
/// @remarks
//---------------------------------------------------------------------------
class empty_mutex_data
{
public:
//------------------------ functions ----------------------
empty_mutex_data ( const empty_mutex_data &) noexcept {};
empty_mutex_data ( void) noexcept {};

empty_mutex_data & operator= ( const empty_mutex_data &) noexcept { return *this ;};
//
//**************************************************************************
//         L O C K  /  U N L O C K     M O D I F Y
//**************************************************************************
//---------------------------------------------------------------------------
//  function : try_lock_write
/// @brief  This function  try to lock for the write operation.
/// @return true : locked     false : unlocked
//---------------------------------------------------------------------------
inline bool try_lock_write ( void) noexcept  { return true ;}

//---------------------------------------------------------------------------
//  function : lock_write
/// @brief  This function  lock for the write operation.
//---------------------------------------------------------------------------
inline void lock_write ( void)  noexcept { };
//---------------------------------------------------------------------------
//  function : unlock_write
/// @brief this function unlock for the write operation
//---------------------------------------------------------------------------
inline void unlock_write( void) noexcept { };
//
//***************************************************************************
//          L O C K  /  U N L O C K     R E A D
//***************************************************************************
//---------------------------------------------------------------------------
//  function : try_lock_read
/// @brief  This function try to lock for the read operation.
/// @return true : locked     false : unlocked
//---------------------------------------------------------------------------
inline bool try_lock_read ( void) noexcept { return true;};
//---------------------------------------------------------------------------
//  function : lock_read
/// @brief  This function lock for the read operation.
//---------------------------------------------------------------------------
inline void lock_read ( void) noexcept { };
//---------------------------------------------------------------------------
//  function :  unlock_read
/// @brief this function unlock for the read operation
//---------------------------------------------------------------------------
inline void unlock_read ( void) noexcept {}

//---------------------------------------------------------------------------
//  function : no_readers
/// @brief This function return true when the number of functions with
///        lock_read reading over the data structure is zero
/// @return true: no readers false : readers
//---------------------------------------------------------------------------
inline bool no_readers ( void) const  noexcept { return true ;};
//---------------------------------------------------------------------------
//  function : n_readers
/// @brief This function return the number of functions with lock_read  reading
///         over the data structure
/// @return number
//---------------------------------------------------------------------------
inline uint32_t n_readers ( void) const  noexcept { return 0 ;};
//---------------------------------------------------------------------------
//  function : wait_no_readers
/// @brief This function wait until the number of readers over the data
///         structure is 0
//---------------------------------------------------------------------------
inline void wait_no_readers () noexcept  {};
//---------------------------------------------------------------------------
};//                end of class empty_mutex_data
//--------------------------------------------------------------------------
//
//###########################################################################
//                                                                         ##
//    ################################################################     ##
//    #                                                              #     ##
//    #                      C L A S S                               #     ##
//    #                 M U T E X _ D A T A                          #     ##
//    #                                                              #     ##
//    ################################################################     ##
//                                                                         ##
//###########################################################################
//
//---------------------------------------------------------------------------
/// @class  mutex_data
/// @brief  This class implements information for to be used by
///         fastmutex_read_cnc and fastmutex_write_cnc in order to block the code\n
/// @remarks
//---------------------------------------------------------------------------
class mutex_data
{
private:
//------------------------- variables --------------------
mutable spinlock spl ;
uint32_t nmutex;
uint32_t nread ;
std::thread::id tid ;

public:
//------------------------ functions ----------------------
inline mutex_data ( const mutex_data &) noexcept :nmutex(0),nread(0){};
inline mutex_data ( void) noexcept :nmutex(0),nread(0){};

inline mutex_data & operator= ( const mutex_data &) noexcept
{   return *this ;
};
inline static std::thread::id this_id ( void) noexcept
{   return std::this_thread::get_id();
};
//
//**************************************************************************
//         L O C K  /  U N L O C K     M O D I F Y
//**************************************************************************
//---------------------------------------------------------------------------
//  function : try_lock_write
/// @brief  This function lock for the write operation. It's atomic
/// @return true : locked     false : unlocked
//---------------------------------------------------------------------------
inline bool try_lock_write ( void) noexcept
{   //---------------------------------- begin --------------------
    std::unique_lock <spinlock> UL ( spl);
    bool SW = ( nmutex == 0 or tid == this_id() );
    if ( SW)
    {   nmutex++ ;
        tid = this_id() ;
    };
    return SW ;
}
//---------------------------------------------------------------------------
//  function : lock_write
/// @brief  This function lock for the write operation. It's atomic
/// @return true : locked     false : unlocked
//---------------------------------------------------------------------------
inline void lock_write(void) noexcept
{   while ( not try_lock_write() )
        std::this_thread::yield();
};
//---------------------------------------------------------------------------
//  function : unlock_write
/// @brief this function unlock the write operation
//---------------------------------------------------------------------------
inline void unlock_write( void) noexcept
{   //-------------------------- begin --------------------
    std::unique_lock <spinlock> UL ( spl);
    assert ( nread== 0 and tid == this_id() and nmutex > 0)  ;
    nmutex --    ;
};
//
//***************************************************************************
//          L O C K  /  U N L O C K     R E A D
//***************************************************************************
//---------------------------------------------------------------------------
//  function : try_lock_read
/// @brief  This function lock for the read operation. It's atomic
/// @param [in]
/// @return true : locked     false : unlocked
//---------------------------------------------------------------------------
inline bool try_lock_read ( void) noexcept
{   //---------------------------------- begin --------------------
    std::unique_lock <spinlock> UL ( spl);
    bool SW = ( nmutex == 0 or tid == this_id ());
    if ( SW)   nread++ ;
    return SW ;
};
//---------------------------------------------------------------------------
//  function : lock_read
/// @brief  This function lock for the read operation. It's atomic
/// @param [in]
/// @return true : locked     false : unlocked
//---------------------------------------------------------------------------
inline void lock_read(void) noexcept
{   //--------------------- begin ----------------------------
    while ( not try_lock_read() )
        std::this_thread::yield();
};
//---------------------------------------------------------------------------
//  function :  unlock_read
/// @brief This function unlock the read operation
//---------------------------------------------------------------------------
inline void unlock_read ( void) noexcept
{   //-------------------------- begin --------------------
    std::unique_lock <spinlock> UL ( spl);
    if ( not( nread > 0 and ( nmutex == 0 or tid == this_id () )))

    assert ( nread > 0 );
    nread--;
};
//---------------------------------------------------------------------------
//  function : no_readers
/// @brief This function return true when the number of functions with
///        lock_read reading over the data structure is zero
/// @return true: no readers false : readers
//---------------------------------------------------------------------------
inline bool no_readers ( void) const noexcept
{   std::unique_lock <spinlock> UL ( spl);
    assert ( nmutex > 0);
    return ( nread == 0) ;
};
//---------------------------------------------------------------------------
//  function : n_readers
/// @brief This function return the number of functions with lock_read  reading
///         over the data structure
/// @return number
//---------------------------------------------------------------------------
inline uint32_t n_readers ( void) const noexcept
{   //-------------------------- begin --------------------
    std::unique_lock <spinlock> UL ( spl);
    assert ( nmutex > 0);
    return nread ;
};
//---------------------------------------------------------------------------
//  function : wait_no_readers
/// @brief This function wait until the number of readers over the data
///         structure is 0
//---------------------------------------------------------------------------
inline void wait_no_readers () noexcept
{   while (not no_readers()) std::this_thread::yield() ;
};
//---------------------------------------------------------------------------
};//                 end of class mutex_data
//---------------------------------------------------------------------------
//
//###########################################################################
//                                                                         ##
//    ################################################################     ##
//    #                                                              #     ##
//    #                        S T R U C T                           #     ##
//    #             C O N F I G _ F A S T M U T E X                  #     ##
//    #                                                              #     ##
//    ################################################################     ##
//                                                                         ##
//###########################################################################
//
//---------------------------------------------------------------------------
/// @class  config_fastmutex
/// @brief  This class implements the fastmutex  selector depending of the
///         concurrency ( cnc) of the data structure\n
//
/// @remarks
//---------------------------------------------------------------------------
template <bool is_concurrent> struct config_fastmutex ;

template < >  struct config_fastmutex <true>
{   typedef mutex_data      fastmutex_data ;
};
template < > struct config_fastmutex < false>
{   typedef empty_mutex_data      fastmutex_data ;
};

//
//###########################################################################
//                                                                         ##
//    ################################################################     ##
//    #                                                              #     ##
//    #                       C L A S S                              #     ##
//    #               F A S T M U T E X _ R E A D                    #     ##
//    #                                                              #     ##
//    ################################################################     ##
//                                                                         ##
//###########################################################################
//
//---------------------------------------------------------------------------
/// @class  mutex_read
/// @brief  This class implements the mutex for the function
///         which read the vector_tree\n
/// @remarks
//---------------------------------------------------------------------------
template <class mutex_data>
class mutex_read
{private :
    //--------------- Variables ---------------
    mutex_data &r ;
    bool locked ;

    //------------- Forbidden Functions -----------------------------
    mutex_read() = delete ;
    mutex_read ( const mutex_read &) = delete ;
    mutex_read & operator = ( const mutex_read &) = delete ;

 public :
    //------------ Public Functions ------------------------
    inline mutex_read( mutex_data & mdc, bool lock_now = true) noexcept
    :r(mdc),locked(false)
    {   if ( lock_now) lock();
    };

    inline ~mutex_read ( void)
    {   if ( locked ) unlock() ;
    };

    inline bool try_lock ( void)
    {   //assert ( not locked) ;
        return (locked = r.try_lock_read()) ;
    };

    inline void lock ( void )
    {   //assert ( not locked) ;
        r.lock_read() ;
        locked = true ;
    };

    inline void unlock ( void)
    {   //assert ( locked );
        r.unlock_read();
        locked = false ;
    };

//---------------------------------------------------------------------------
};//                 End of class mutex_read
//---------------------------------------------------------------------------
//
//###########################################################################
//                                                                         ##
//    ################################################################     ##
//    #                                                              #     ##
//    #                     C L A S S                                #     ##
//    #              M U T E X _ W R I T E                           #     ##
//    #                                                              #     ##
//    ################################################################     ##
//                                                                         ##
//###########################################################################
//
//-------------------------------------------------------------
/// @class  mutex_write
/// @brief  This class implements the mutex for the function
///         which write the vector_tree_cnc\n
/// @remarks
//----------------------------------------------------------------
template <class mutex_data>
class mutex_write
{private :
    //----------------- Variables ---------------------
    mutex_data &r ;
    bool locked ;

    //----------- Forbidden Functions ----------------------------
    mutex_write() = delete ;
    mutex_write ( const mutex_write &) = delete ;
    mutex_write & operator = ( const mutex_write &) = delete ;

 public :
    inline mutex_write (mutex_data & mm , bool lock_now = true) noexcept :
        r(mm),locked (false)
    {   if ( lock_now) lock();
    };

    inline ~mutex_write ( void) noexcept
    {   if ( locked) unlock();
    };

    inline bool try_lock ( void) noexcept
    {   //assert ( not locked) ;
        return ( locked= r.try_lock_write() );
    };

    inline void lock ( void) noexcept
    {   //assert ( not locked);
        r.lock_write();
        locked = true ;
    };

    inline void unlock( void) noexcept
    {   //assert ( locked) ;
        r.unlock_write();
        locked = false ;
    };

    inline void wait_no_readers () noexcept
    {   //assert ( locked) ;
        r.wait_no_readers() ;
    };
//---------------------------------------------------------------------------
};//                   end of class mutex_write
//---------------------------------------------------------------------------

//
//###########################################################################
//                                                                         ##
//    ################################################################     ##
//    #                                                              #     ##
//    #                        C L A S S                             #     ##
//    #              M U T E X _ W R I T E _ R E A D                 #     ##
//    #                                                              #     ##
//    ################################################################     ##
//                                                                         ##
//###########################################################################
//
//---------------------------------------------------------------------------
/// @class  mutex_write_read
/// @brief  This class implements information for to be used by
///         mutex_read_empty and mutex_write_empty in order to block the code\n
/// @remarks
//---------------------------------------------------------------------------
template <class mutex_data1 , class mutex_data2>
class mutex_write_read
{private :
    mutex_write<mutex_data1> mtx_w ;
    mutex_read<mutex_data2>  mtx_r ;
    bool equal ;
    bool locked ;

    //------------- Forbidden Functions -------------------------------
    mutex_write_read() = delete ;
    mutex_write_read ( const mutex_write_read &) = delete ;
    mutex_write_read & operator = ( const mutex_write_read &) = delete ;

 public :
    //------------ Public Functions -------------------------
    inline mutex_write_read ( mutex_data1 &mw,mutex_data2 &mr, bool lock_now=true) noexcept
                       : mtx_w(mw,false), mtx_r(mr,false),locked(false)
    {   equal = ((void*) (&mw) == (void*) (&mr));
        if ( lock_now) lock() ;
    };
    inline ~mutex_write_read ( void) noexcept
    {   if ( locked) unlock() ;
    };

    inline bool try_lock ( void) noexcept
    {   assert ( not locked);
        return (locked=(equal)? mtx_w.try_lock():std::try_lock(mtx_w,mtx_r)== -1);
    };

    inline void lock ( void) noexcept
    {   //assert ( not locked);
        (equal)? mtx_w.lock(): std::lock(  mtx_w , mtx_r );
        locked = true ;
    };

    inline void unlock ( void) noexcept
    {   //assert ( locked);
        mtx_w.unlock() ;
        if ( not equal) mtx_r.unlock() ;
        locked = false ;
    };

    inline void wait_no_readers () noexcept
    {   //assert ( locked) ;
        mtx_w.wait_no_readers() ;
    };

    inline mutex_write<mutex_data1>& mtx_write ( void) noexcept { return mtx_w; } ;

    inline mutex_read<mutex_data2> & mtr_read ( void) noexcept  { return mtx_r ;};
//---------------------------------------------------------------------------
};//              End of class mutex_write_read
//---------------------------------------------------------------------------
//
//###########################################################################
//                                                                         ##
//    ################################################################     ##
//    #                                                              #     ##
//    #                      C L A S S                               #     ##
//    #            M U T E X _ W R I T E _ W R I T E                 #     ##
//    #                                                              #     ##
//    ################################################################     ##
//                                                                         ##
//###########################################################################
//
//---------------------------------------------------------------------------
/// @class  mutex_write_write
/// @brief  This class implements information for to be used by
///         mutex_read_empty and mutex_write_empty in order to block the code\n
/// @remarks
//---------------------------------------------------------------------------
template <class mutex_data1 ,class mutex_data2>
class mutex_write_write
{private :
    mutex_write<mutex_data1> mtx_w1 ;
    mutex_write<mutex_data2> mtx_w2 ;
    bool equal , locked;

    //------------- Forbidden Functions -------------------------------
    mutex_write_write() = delete ;
    mutex_write_write ( const mutex_write_write &) = delete ;
    mutex_write_write & operator = ( const mutex_write_write &) = delete ;

 public :
    //------------ Public Functions -------------------------
    inline mutex_write_write( mutex_data1 &mw1 ,mutex_data2 &mw2, bool lock_now=true ) noexcept
                       :mtx_w1(mw1,false),mtx_w2(mw2,false),locked ( false)
    {   //----------------------------- begin --------------------------------
        equal = ((void*) (&mw1) == (void*) (&mw2));
        if ( lock_now) lock() ;
    };
    inline ~mutex_write_write (void) noexcept
    {   if(locked) unlock();
    };
    inline bool try_lock ( void ) noexcept
    {   //assert ( not locked);
        return ( locked = (equal)? mtx_w1.try_lock() :std::try_lock (mtx_w1, mtx_w2 ) == -1);
    };
    inline void lock ( void) noexcept
    {   //assert ( not locked );
        if (equal) mtx_w1.lock();
        else std::lock(mtx_w1,mtx_w2 );
        locked =true ;
    };

    inline void unlock ( void) noexcept
    {   //assert( locked);
        mtx_w1.unlock() ;
        if ( not equal) mtx_w2.unlock() ;
        locked =false ;
    };
    inline mutex_write<mutex_data1>& mtx_write_first  ( void) noexcept  { return mtx_w1; } ;

    inline mutex_read<mutex_data2> & mtr_write_second ( void) noexcept  { return mtx_w2 ;};

    inline void wait_no_readers_first  () noexcept
    {   //assert (locked);
        mtx_w1.wait_no_readers() ;
    };

    inline void wait_no_readers_second () noexcept
    {   //assert (locked);
        if ( equal) mtx_w1.wait_no_readers() ;
        else        mtx_w2.wait_no_readers() ;
    };
//---------------------------------------------------------------------------
};//              End of class mutex_write_write
//---------------------------------------------------------------------------
//
//###########################################################################
//                                                                         ##
//    ################################################################     ##
//    #                                                              #     ##
//    #                        C L A S S                             #     ##
//    #                M U T E X _ R E A D _ R E A D                 #     ##
//    #                                                              #     ##
//    ################################################################     ##
//                                                                         ##
//###########################################################################
//
//---------------------------------------------------------------------------
/// @class  mutex_read_read
/// @brief  This class implements information for to be used by
///         mutex_read_empty and mutex_write_empty in order to block the code\n
/// @remarks
//---------------------------------------------------------------------------
template <class mutex_data1 , class mutex_data2>
class mutex_read_read
{private :
    mutex_read<mutex_data1>  mtx_r1 ;
    mutex_read<mutex_data1>  mtx_r2 ;
    bool equal,locked ;

    //------------- Forbidden Functions -------------------------------
    mutex_read_read() = delete ;
    mutex_read_read ( const mutex_read_read &) = delete ;
    mutex_read_read & operator = ( const mutex_read_read &) = delete ;

 public :
    //------------ Public Functions -------------------------
    inline mutex_read_read( mutex_data1 &mr1 , mutex_data2 &mr2, bool lock_now=true ) noexcept
    :mtx_r1(mr1,false),mtx_r2(mr2,false),locked(false)
    {   equal = ((void*) (&mr1) == (void*) (&mr2));
        if ( lock_now) lock() ;
    };
    inline ~mutex_read_read( void) noexcept
    {   if ( locked) unlock() ;

    }
    inline bool try_lock ( void) noexcept
    {   //assert ( not locked);
        return ( locked = (equal)? mtx_r1.try_lock() :std::try_lock ( mtx_r1 , mtx_r2 ) == -1);
    };

    inline void lock ( void) noexcept
    {   //assert ( not locked) ;
        (equal)? mtx_r1.lock(): std::lock( mtx_r1 , mtx_r2 );
        locked = true ;
    };

    inline void unlock  ( void) noexcept
    {   //----------------------- begin -------------------------------
        //assert ( locked );
        mtx_r1.unlock();
        if ( equal ) mtx_r2.unlock();
        locked = false ;
    };
//---------------------------------------------------------------------------
};//               End of class mutex_read_read
//---------------------------------------------------------------------------

//***************************************************************************
};//    End namespace util
};//    End namespace countertree
};//    End namespace boost
//***************************************************************************
#endif
