//----------------------------------------------------------------------------
/// @file merge_sort.hpp
/// @brief Quick Sort algorithm
///
/// @author Copyright (c) 2010 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//-----------------------------------------------------------------------------
#ifndef __CNTREE_SORT_PARALLEL_MERGE_SORT_HPP
#define __CNTREE_SORT_PARALLEL_MERGE_SORT_HPP

#include <boost/countertree/util/atomic.hpp>
#include <boost/countertree/util/stack_cnc.hpp>
#include <boost/countertree/util/sort/merge_sort.hpp>
#include <boost/countertree/util/sort/indirect_sort.hpp>
#include <functional>
#include <vector>


namespace boost
{
namespace countertree
{
namespace util
{
namespace sort
{
using namespace boost::countertree::util;

template < class iter_t,
          typename compare = std::less < typename iter_value<iter_t>::type >
        >
struct parallel_merge_sort_comp
{
//-------------------------------------------------------------------------
//                     definitions
//-------------------------------------------------------------------------
typedef typename iter_value<iter_t>::type value_t ;
typedef std::atomic<size_t> count_t;
typedef std::atomic<uint32_t> join_t ;

//-------------------------------------------------------------------------
//                     definitions
//-------------------------------------------------------------------------
static const size_t NELEM_MIN = 1<<16 ;
iter_t global_first, global_last ;
count_t VCount[48] ;
join_t VJoin[48] ;
count_t NSort ;
join_t JoinSort ;
size_t Chuck_size ;
size_t N ;
uint32_t NLevel ;
NThread NT;
compare comp ;


//#########################################################################
//
//                F U N C T I O N S
//
//#########################################################################
parallel_merge_sort_comp (iter_t first, iter_t last ,NThread NT1=NThread())
                         :parallel_merge_sort_comp(first,last,compare(),NT1){};


parallel_merge_sort_comp ( iter_t first, iter_t last,
                           compare comp1, NThread NT1=NThread())
                         : global_first(first), global_last (last),
                           NT ( NT1),comp(comp1)
{   //------------------------------- begin -----------------------------
    auto  N1 = last -first;
    assert ( N1 >= 0 );
    N = (size_t) N1 ;
    if ( N < NELEM_MIN or NT() == 1)
    {   merge_sort ( first , last, comp );
        return ;
    };

    //---------------------------------------------------------------------
    //     Parallel Process
    //---------------------------------------------------------------------
    NLevel = NBits ( NT() ) ;
    if ( (1U<<NLevel) != NT()) NLevel++ ;

    Chuck_size = ( N + NMax(NLevel) -1)/NMax(NLevel);
    //for ( uint32_t i = 0 ; i < (64-BitsMax) ;++i)
    for ( uint32_t i = 0 ; i < NLevel ;++i)
    {   VCount[i].store (  0);
        VJoin[i].store (  0);
    };
    NSort.store ( 0);
    JoinSort.store (0);

    //Para una sola hebra
    //sort_thread() ;

    //---------------------------------------------------------------------
    //     Thread process
    //---------------------------------------------------------------------
    std::vector <std::thread > VT ( NT());

    for ( uint32_t i  =0  ; i < NT() ; ++i)
        VT[i] = std::thread ( &parallel_merge_sort_comp::sort_thread, this );
    for ( uint32_t i  =0  ; i < NT() ; ++i)
        VT[i].join();


}; // end of constructor

inline void merge (iter_t Alfa, value_t *P1, value_t *P1_end,
                                iter_t P2, iter_t P2_end         )
{   //---------------------------- begin -----------------------------------
    //value_t * P = P1 ;
    bool SW ;
    for ( ;P2 != P2_end and Alfa != P2 ;++Alfa  )
    {   //*(Alfa++)= (comp(*P2,*P1))?std::move(*(P2++)):std::move(*(P1++));
        SW =  comp(*P2,*P1);
        *Alfa= ( SW)?std::move(*(P2++)):std::move(*P1);
        if (  not SW) (P1++)->~value_t() ;
    }
    if ( P2 == P2_end)
    {   //for (; P1!= P1_end; ++P1 ) *(Alfa++) =std::move(*P1);
        for (; P1!= P1_end; ++P1 )
        {   *(Alfa++) =std::move(*P1);
            P1->~value_t() ;
        };
    };
    //for ( P1 = P; P1 != P1_end; ++P1) P1->~value_t() ;
};

inline size_t NMax ( uint32_t Level)
{   return ( (size_t)1 )<<Level ;
};

inline void Chuck (uint32_t Level,size_t Pos,iter_t &first,iter_t &last)
{   //-------------------------- begin ----------------------------
    size_t Max = NMax ( Level);
    assert ( Pos < Max);
    size_t Size = Chuck_size <<( NLevel -Level);
    first = global_first + ( Size *Pos);
    last = (Pos == (Max-1))?global_last : first + Size ;
};

void sort_thread( )
{   //------------------------- begin ----------------------
    iter_t first1 , first2 , last1 , last2 ;
    value_t *P = nullptr;
    size_t Size ;
    //---------------------------------------------------------------------
    //                        sorting
    //---------------------------------------------------------------------
    size_t Pos  = 0;
    //while ( (Pos = increment(NSort)) < NMax(NLevel))
    while ( (Pos = atomic_add(NSort,1)) < NMax(NLevel))
    {   Chuck ( NLevel, Pos, first1 , last1);
        Size = (Chuck_size +1) >>1;
        if ( P == nullptr)
        {   P = std::get_temporary_buffer <value_t> (Size).first ;
            if ( P == nullptr) throw std::bad_alloc() ;
        };
        merge_sort_tag<iter_t,compare>::merge_sort_internal( first1,last1,comp,P);
    };
    atomic_add ( JoinSort,1);
    if ( P != nullptr) std::return_temporary_buffer (P) ;
    P = nullptr ;
    while ( atomic_read ( JoinSort) < NT()) std::this_thread::yield() ;

    //---------------------------------------------------------------------
    //                         merging
    //---------------------------------------------------------------------
    for ( uint32_t Level = NLevel ; Level != 0 ; --Level)
    {   Size = Chuck_size <<( NLevel - Level );
        while ((Pos = atomic_add ( VCount[Level-1],1)) < NMax ( Level -1) )
        {   if ( P == nullptr)
            {   P = std::get_temporary_buffer <value_t> (Size).first ;
                if ( P == nullptr) throw std::bad_alloc() ;

                //void * Ptr = ::operator new ( Size * sizeof (value_t));
                //if ( Ptr == nullptr) throw std::bad_alloc ( );
                //P = static_cast <value_t*> ( Ptr);
            };
            Chuck ( Level, Pos<<1, first1, last1);
            Chuck ( Level, (Pos<<1)+1, first2 , last2);
            //move_to_uninitialized_memory ( first1, P,Size);
            iter_t src = first1;
            value_t *dest = P ;
            for ( size_t i =0 ; i < Size ; ++i)
                ::new ( static_cast<void*>((dest++))) value_t( std::move(*(src++)) );

            merge ( first1 , P , P+Size, first2 , last2);
            //destroy ( P , Size);
        };
        atomic_add ( VJoin[Level], 1);
        //if ( P != nullptr)  ::operator delete (static_cast <void*> (P)) ;
        if ( P != nullptr) std::return_temporary_buffer (P) ;
        P = nullptr ;
        while ( atomic_read ( VJoin[Level]) < NT()) std::this_thread::yield() ;
    };
};
};// end struct parallel_merge_sort

//------------------------------------------------------------------------------
// These functions are for to select the correct format depending of the number
// and type of the parameters
//------------------------------------------------------------------------------
template < class iter_t,
          typename compare = std::less < typename iter_value<iter_t>::type >
        >
void parallel_merge_sort ( iter_t first, iter_t last , NThread NT = NThread() )
{   parallel_merge_sort_comp <iter_t> ( first, last, NT);
};

template < class iter_t,
          typename compare = std::less < typename iter_value<iter_t>::type >
        >
void parallel_merge_sort ( iter_t first, iter_t last,
                           compare comp1, NThread NT = NThread() )
{   //----------------------------- begin ----------------------------------
    parallel_merge_sort_comp<iter_t,compare> ( first, last,comp1,NT);
};

template < class iter_t,
          typename compare = std::less < typename iter_value<iter_t>::type >
        >
void parallel_stable_sort ( iter_t first, iter_t last , NThread NT = NThread() )
{   parallel_merge_sort_comp <iter_t> ( first, last, NT);
};

template < class iter_t,
          typename compare = std::less < typename iter_value<iter_t>::type >
        >
void parallel_stable_sort ( iter_t first, iter_t last,
                            compare comp1 , NThread NT = NThread())
{   //----------------------------- begin -----------------------------------
    parallel_merge_sort_comp<iter_t,compare> ( first, last,comp1,NT);
};
template < class iter_t,
          typename compare = std::less < typename iter_value<iter_t>::type >
        >
void indirect_parallel_stable_sort ( iter_t first, iter_t last , NThread NT = NThread() )
{  	parallel_indirect_sort_method<parallel_merge_sort_comp, iter_t,compare> ( first , last, compare() );
};
template < class iter_t,
          typename compare = std::less < typename iter_value<iter_t>::type >
        >
void indirect_parallel_merge_sort ( iter_t first, iter_t last , NThread NT = NThread() )
{  	parallel_indirect_sort_method<parallel_merge_sort_comp, iter_t,compare> ( first , last, compare() );
};
/*
//---------------------------------------------------------------------------
//  template <class iter_t  >
//  void parallel_merge_sort ( iter_t first, iter_t last,
//                             uint32_t NThread = NThread_HW );
//
//  template <class iter_t , typename compare  >
//  void parallel_merge_sort ( iter_t first, iter_t last,
//                             compare comp1 , uint32_t NThread= NThread_HW);
//
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
// This struct is for to capture the number of threads from any
// integral format passed as third parameter
// If the third paramenter is not integral it's a comparator
//----------------------------------------------------------------------------
template < class iter_t,typename compare ,bool SW> struct selector_merge_sort ;

template <class iter_t , typename compare >
struct selector_merge_sort<iter_t,compare,false>
{   selector_merge_sort( iter_t first, iter_t last, compare comp1  )
    {   parallel_merge_sort_comp<iter_t,compare> ( first,last,comp1);
    };
};

template <class iter_t , typename Aux   >
struct selector_merge_sort<iter_t,Aux,true>
{   selector_merge_sort ( iter_t first, iter_t last, Aux NThread )
    {   parallel_merge_sort_comp<iter_t> (first, last,NThread);
    };
};

//------------------------------------------------------------------------------
// These functions are for to select the correct format depending of the number
// and type of the parameters
//------------------------------------------------------------------------------
template <class iter_t  >
void parallel_merge_sort ( iter_t first, iter_t last  )
{   parallel_merge_sort_comp <iter_t> ( first, last);
};

template <class iter_t , typename compare   >
void parallel_merge_sort ( iter_t first, iter_t last, compare comp1 , uint32_t NThread )
{   parallel_merge_sort_comp<iter_t,compare> ( first, last,comp1,NThread);
};

template <class iter_t , typename compare   >
void parallel_merge_sort ( iter_t first, iter_t last, compare comp1 )
{   selector_merge_sort<iter_t,compare,std::is_integral<compare>::value> ( first, last,comp1);
};
*/
//
//****************************************************************************
};//    End namespace sort
};//    End namespace util
};//    End namespace countertree
};//    End namespace boost
//****************************************************************************
//
#endif
