//----------------------------------------------------------------------------
/// @file spinlock.hpp
/// @brief
///
/// @author Copyright (c) 2010 2013 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//-----------------------------------------------------------------------------
#ifndef __CNTREE_UTIL_SPINLOCK_HPP
#define __CNTREE_UTIL_SPINLOCK_HPP

#include <atomic>
#include <memory>
#include <functional>
#include <mutex>
#include <thread>
#include <ctime>

namespace boost
{
namespace countertree
{
namespace util
{

//##########################################################################
//                                                                        ##
//                 C L A S S     S P I N L O C K                          ##
//                                                                        ##
//##########################################################################
//---------------------------------------------------------------------------
/// @class spinlock
///
/// @remarks This class meet the BasicLockable requirements ( lock, unlock )
//---------------------------------------------------------------------------
class spinlock
{
private:
//***************************************************************************
//             P R I V A T E      V A R I A B L E S
//***************************************************************************
std::atomic_flag af = ATOMIC_FLAG_INIT;


public :
//---------------------------------------------------------------------------
//  function : spinlock
/// @brief  class constructor
/// @param [in]
//---------------------------------------------------------------------------
explicit inline spinlock ( ) noexcept { };

//---------------------------------------------------------------------------
//  function : lock
/// @brief  Lock the spinlock
//---------------------------------------------------------------------------
inline void lock() noexcept
{   //while ( af.test_and_set(std::memory_order_acquire) and wait());
    if ( af.test_and_set(std::memory_order_acquire))
    {   while ( wait() and  af.test_and_set(std::memory_order_relaxed) );
    };
};

//---------------------------------------------------------------------------
//  function : wait
/// @brief
/// @param [in]
/// @return true :locked false: not previous locked
//---------------------------------------------------------------------------
inline bool wait () const noexcept
{   std::this_thread::yield();
    return true ;
} ;

//---------------------------------------------------------------------------
//  function : try_lock
/// @brief Try to lock the spinlock, if not, return false
/// @param [in]
/// @return true :locked false: not previous locked
//---------------------------------------------------------------------------
inline bool try_lock() noexcept
{   return not af.test_and_set(std::memory_order_acquire);
};

//---------------------------------------------------------------------------
//  function : unlock
/// @brief  unlock the spinlock
//---------------------------------------------------------------------------
inline void unlock() noexcept
{   //----------------------- begin -----------------
    af.clear(std::memory_order_release );
};

//***************************************************************************
};// E N D    C L A S S     S P I N L O C K
//***************************************************************************

//##########################################################################
//                                                                        ##
//               C L A S S     N U L L _ S P I N L O C K                  ##
//                                                                        ##
//##########################################################################
//---------------------------------------------------------------------------
/// @class null_spinlock
///
/// @remarks
//---------------------------------------------------------------------------
class null_spinlock
{

public :
//---------------------------------------------------------------------------
//  function : null_spinlock
/// @brief  class constructor
/// @param [in]
//---------------------------------------------------------------------------
explicit inline null_spinlock ( ) noexcept { };

//---------------------------------------------------------------------------
//  function : lock
/// @brief  Lock the spinlock
//---------------------------------------------------------------------------
inline void lock() noexcept {};

//---------------------------------------------------------------------------
//  function : wait
/// @brief
/// @param [in]
/// @return true :locked false: not previous locked
//---------------------------------------------------------------------------
inline bool wait () const noexcept { return true ;} ;

//---------------------------------------------------------------------------
//  function : try_lock
/// @brief Try to lock the spinlock, if not, return false
/// @param [in]
/// @return true :locked false: not previous locked
//---------------------------------------------------------------------------
inline bool try_lock() noexcept { return true;};

//---------------------------------------------------------------------------
//  function : unlock
/// @brief  unlock the spinlock
//---------------------------------------------------------------------------
inline void unlock() noexcept { };

//***************************************************************************
};// E N D    C L A S S     N U L L _ S P I N L O C K
//***************************************************************************

//##########################################################################
//                                                                        ##
//    C L A S S     F A S T _ L O C K < C O N C R R E N T >               ##
//                                                                        ##
//##########################################################################
//---------------------------------------------------------------------------
/// @class config_lock
/// @brief : Template for to select the lock depending if is concurrent
///          or not
///
/// @remarks
//---------------------------------------------------------------------------
template < bool concurrent >
struct fastmutex ;

template < >
struct fastmutex <true>
{   typedef spinlock type ;
};
template < >
struct fastmutex <false>
{   typedef null_spinlock   type ;
};

/*
//##########################################################################
//                                                                        ##
//            C L A S S      L O C K _ D A T A _ C N C                    ##
//                                                                        ##
//##########################################################################
//-------------------------------------------------------------
/// @class  lock_data_cnc
/// @brief  This class implements the data  for the lock
//
/// @remarks
//----------------------------------------------------------------
template <class splock = spinlock  >
struct lock_data_cnc
{   splock SL ;
};

//##########################################################################
//                                                                        ##
//             C L A S S        L O C K _ C N C                           ##
//                                                                        ##
//##########################################################################

//-------------------------------------------------------------
/// @class  lock_cnc
/// @brief  This class implements the barrier for the function
///         which modify the vector_tree_cnc_cnc\n
//
/// @remarks
//----------------------------------------------------------------
template <class splock = spinlock  >
class lock_cnc
{private :
    //-------------- Variables ---------------------------
    std::unique_lock <splock> UL ;

    //------------ Forbidden Functions -------------------
    lock_cnc() = delete ;
    lock_cnc ( const lock_cnc &) = delete ;
    lock_cnc & operator = ( const lock_cnc &) = delete ;

 public :
    inline lock_cnc (lock_data_cnc<splock> & BDC):UL(BDC.SL){} ;
    inline ~lock_cnc() {  };
};

//##########################################################################
//                                                                        ##
//         C L A S S        L O C K _ D A T A _ E M P T Y                 ##
//                                                                        ##
//##########################################################################
//-------------------------------------------------------------
/// @class  lock_data_empty
/// @brief  This class implements the data  for the lock
//
/// @remarks
//----------------------------------------------------------------
struct lock_data_empty {  };


//##########################################################################
//                                                                        ##
//             C L A S S        L O C K _ E M P T Y                       ##
//                                                                        ##
//##########################################################################
//-------------------------------------------------------------
/// @class  lock_empty
/// @brief  This class implements the barrier for the function
///         which modify the vector_tree_cnc_cnc\n
//
/// @remarks
//----------------------------------------------------------------
class lock_empty
{private :
    //------------------- Forbidden Functions --------------------
    lock_empty() = delete ;
    lock_empty ( const lock_empty &) = delete ;
    lock_empty & operator = ( const lock_empty &) = delete ;

 public :
    inline lock_empty (lock_data_empty &){} ;
    inline ~lock_empty() {  };
};

//##########################################################################
//                                                                        ##
//    C L A S S   C O N F I G _ L O C K < I S _ C O N C U R R E N T  >    ##
//                                                                        ##
//##########################################################################
//---------------------------------------------------------------------------
/// @class config_lock
/// @brief : Template for to select the lock depending if is concurrent
///          or not
///
/// @remarks
//---------------------------------------------------------------------------
template < class splock ,bool is_concurrent >
struct config_lock ;

template <class splock >
struct config_lock <splock ,true>
{   typedef lock_data_cnc<splock>    lock_data ;
    typedef lock_cnc<splock>         mylock ;
};
template < class splock>
struct config_lock <splock, false>
{   typedef lock_data_empty         lock_data ;
    typedef lock_empty              mylock ;
};
*/
//***************************************************************************
};// end namespace util
};// end namespace countertree
};// end namespace boost
//***************************************************************************
#endif
