//----------------------------------------------------------------------------
/// @file util_iterator.hpp
/// @brief Utility functions for the iterators
/// @author Copyright (c) 2013 Francisco José Tapia (fjtapia@gmail.com )\n
///         Distributed under the Boost Software License, Version 1.0.\n
///         ( See accompanyingfile LICENSE_1_0.txt or copy at
///           http://www.boost.org/LICENSE_1_0.txt  )
/// @version 0.1
///
/// @remarks
//
//  template <typename iter_t>
//  struct iter_value
//
//  template <typename iter_t>
//  struct iter_random
//
//  template <class iter_t >
//  struct iter_ptr
//
//  template <class T , class comp_t = std::less<T> >
//  struct less_ptr
//
//  template <class T , class comp_t = std::less<T> >
//  struct less_ptr_no_null
//
//-----------------------------------------------------------------------------
#ifndef __CNTREE_UTIL_UTIL_ITERATOR_HPP
#define __CNTREE_UTIL_UTIL_ITERATOR_HPP

#include <cassert>
#include <utility>
#include <type_traits>
#include <iterator>
#include <functional>
#include <boost/countertree/util/config.hpp>
#include <boost/countertree/util/definition.hpp>

namespace boost
{
namespace countertree
{
namespace util
{
//
//############################################################################
//                                                                          ##
//                         T R A I T S                                      ##
//                                                                          ##
//                  M E T A P R O G R A M M I N G                           ##
//                                                                          ##
//############################################################################
//---------------------------------------------------------------------------
//         D E F I N I T I O N S
//---------------------------------------------------------------------------

template <typename iter_t>
struct iter_value
{   typedef typename
    std::remove_reference<decltype(*(std::declval<iter_t>()))>::type type ;
};

template <typename iter_t>
using iter_random =
typename std::is_same <typename std::iterator_traits<iter_t>::iterator_category,
                       std::random_access_iterator_tag>  ;

template <typename iter_t>
using  enable_if_random =
typename std::enable_if<iter_random<iter_t>::value >::type;

template <typename iter_t>
using  enable_if_not_random =
typename std::enable_if<not iter_random<iter_t>::value >::type;



//############################################################################
//                                                                          ##
//                          I T E R A T O R                                 ##
//                                                                          ##
//                       E N V E L O P M E N T S                            ##
//                                                                          ##
//############################################################################

//
//##########################################################################
//                                                                        ##
//                    C L A S S     I T E R _ P T R                       ##
//                                                                        ##
//##########################################################################
///---------------------------------------------------------------------------
/// @struct iter_ptr
/// @brief This is an iterator useful when the pointed by the iter_t iterators
///        are pointers. And you need to sort the pointers according the values
///        pointed by the operator * of iter_t
/// @tparam iter_t : iterator, must point to pointers
/// @remarks
//----------------------------------------------------------------------------
template <class iter_t >
struct iter_ptr
{   //---------------------- static_assert -----------------------------------
    static_assert ( std::is_pointer <typename iter_value<iter_t>::type>::value,
                   "The iter_t must return a pointer") ;
    typedef std::random_access_iterator_tag             iterator_category   ;
    typedef typename iter_value<iter_t>::type value_t ;
    //------------------- variables ------------------------------------------
    iter_t it ;

    //--------------------- functions -----------------------
    iter_ptr ( iter_t it2) : it ( it2){};
    iter_ptr ( const iter_ptr &it3):it ( it3.it){};

    iter_ptr & operator = ( iter_t it2 ){ it = it2; return *this ;};
    iter_ptr & operator = ( iter_ptr it3){ it = it3.it; return *this ;};

    bool operator == ( iter_ptr it3)const{ return ( it == it3.it) ;};
    bool operator != ( iter_ptr it3)const{ return ( it != it3.it) ;};
    bool operator <  ( iter_ptr it3)const{ return ( it <  it3.it) ;};
    bool operator >  ( iter_ptr it3)const{ return ( it >  it3.it) ;};
    bool operator <= ( iter_ptr it3)const{ return ( it <= it3.it) ;};
    bool operator >= ( iter_ptr it3)const{ return ( it >= it3.it) ;};

    iter_ptr & operator ++ ( void) { ++it ; return *this ;};
    iter_ptr & operator -- ( void) { --it ; return *this ;};

    iter_ptr & operator += ( signed_type N){ it+= N ; return *this ;};
    iter_ptr & operator -= ( signed_type N){ it-= N ; return *this ;};

    iter_ptr operator + ( signed_type N )const{ return (it + N );};
    iter_ptr operator - ( signed_type N )const{ return (it - N );};

    signed_type operator - ( iter_ptr it3) const { return (it- it3.it);};

    value_t &  operator * ( void){ return (*( *it) );};
    const value_t &  operator * ( void)const { return (*( *it) );};

    value_t &  operator -> ( void){ return (*( *it) );};
    const value_t &  operator -> ( void)const { return (*( *it) );};
};
///---------------------------------------------------------------------------
/// @struct iter_rvalue
/// @brief This is an iterator useful when the pointed by the iter_t iterators
///        are pointers. And you need to sort the pointers according the values
///        pointed by the operator * of iter_t
/// @tparam iter_t : iterator, must point to pointers
/// @remarks
//----------------------------------------------------------------------------
template <class iter_t >
struct iter_rvalue
{   //---------------------- static_assert -----------------------------------
    static_assert ( std::is_pointer <typename iter_value<iter_t>::type>::value,
                   "The iter_t must return a pointer") ;
    typedef std::random_access_iterator_tag             iterator_category   ;
        typedef typename iter_value<iter_t>::type value_t ;
    //------------------- variables ------------------------------------------
    iter_t it ;

    //--------------------- functions -----------------------
    iter_rvalue ( iter_t it2) : it ( it2){};
    iter_rvalue ( const iter_rvalue &it3):it ( it3.it){};

    iter_rvalue & operator = ( iter_t it2 ){ it = it2; return *this ;};
    iter_rvalue & operator = ( iter_rvalue it3){ it = it3.it; return *this ;};

    bool operator == ( iter_rvalue it3)const{ return ( it == it3.it) ;};
    bool operator != ( iter_rvalue it3)const{ return ( it != it3.it) ;};
    bool operator <  ( iter_rvalue it3)const{ return ( it <  it3.it) ;};
    bool operator >  ( iter_rvalue it3)const{ return ( it >  it3.it) ;};
    bool operator <= ( iter_rvalue it3)const{ return ( it <= it3.it) ;};
    bool operator >= ( iter_rvalue it3)const{ return ( it >= it3.it) ;};

    iter_rvalue & operator ++ ( void) { ++it ; return *this ;};
    iter_rvalue & operator -- ( void) { --it ; return *this ;};

    iter_rvalue & operator += ( signed_type N){ it+= N ; return *this ;};
    iter_rvalue & operator -= ( signed_type N){ it-= N ; return *this ;};

    iter_rvalue operator + ( signed_type N )const{ return (it + N );};
    iter_rvalue operator - ( signed_type N )const{ return (it - N );};

    signed_type operator - ( iter_rvalue it3) const { return (it- it3.it);};

    value_t &&  operator * ( void){ return (*( *it) );};
    const value_t &&  operator * ( void)const { return (*( *it) );};

    value_t &&  operator -> ( void){ return std::move( *it) ;};
    const value_t &&  operator -> ( void)const { return std::move( *it);};
};
//---------------------------------------------------------------------------
/// @struct iter_const
/// @brief This an iterator which provide always the same value
/// @tparam value_t : type of the value returned
/// @remarks This is useful for to fill the data structure with the same value
//----------------------------------------------------------------------------
template <class value_t >
struct iter_const
{   //------------------------------- variables -------------------------------
    size_type Pos ;
    value_t value ;
    typedef std::random_access_iterator_tag             iterator_category   ;
    //--------------------------- functions ----------------------------------
    iter_const ( size_type Ps=0 ,const value_t &val = value_t ()):Pos(Ps),value (val){};
    iter_const ( const iter_const &it):Pos ( it.Pos), value(it.value){};

    iter_const & operator = ( iter_const it)
    {   Pos = it.Pos;
        value = it.value ;
        return *this ;
    };

    bool operator == ( iter_const it)const{ return  Pos == it.Pos ;};
    bool operator != ( iter_const it)const{ return  Pos != it.Pos  ;};
    bool operator <  ( iter_const it)const{ return  Pos < it.Pos  ;};
    bool operator >  ( iter_const it)const{ return  Pos > it.Pos  ;};
    bool operator <= ( iter_const it)const{ return  Pos <= it.Pos  ;};
    bool operator >= ( iter_const it)const{ return  Pos >= it.Pos  ;};

    iter_const & operator ++ ( void) { ++Pos ; return *this ;};
    iter_const & operator -- ( void) { --Pos ; return *this ;};

    iter_const & operator += ( signed_type N){ Pos+= N ; return *this ;};
    iter_const & operator -= ( signed_type N){ Pos-= N ; return *this ;};

    iter_const operator + ( signed_type N )const
    {   return iter_const (Pos + N, value );
    };
    iter_const operator - ( signed_type N )const
    {   return iter_const (Pos - N, value );
    };
    signed_type operator - ( iter_const it) const { return (Pos- it.Pos);};

    value_t &  operator * ( void){ return value;};
    const value_t &  operator * ( void)const { return value;};

    value_t &  operator -> ( void){ return value;};
    const value_t &  operator -> ( void)const { return value;};
};
//---------------------------------------------------------------------------
/// @struct iter_function
/// @brief This an iterator which provide always the same value
/// @tparam value_t : type of the value returned
/// @remarks This is useful for to fill the data structure with the result
///          of a function
//----------------------------------------------------------------------------
template <class value_t ,class function_t >
struct iter_function
{   //------------------------------- variables -------------------------------
    size_type Pos ;
    function_t function ;
    typedef std::random_access_iterator_tag             iterator_category   ;
    //--------------------------- functions ----------------------------------
    iter_function ( size_type Ps ,const function_t &fnc):Pos(Ps),function (fnc){};
    iter_function ( const iter_function &it):Pos ( it.Pos), function(it.function){};

    iter_function & operator = ( iter_function it)
    {   Pos = it.Pos;
        function = it.function ;
        return *this ;
    };

    bool operator == ( iter_function it)const{ return  Pos == it.Pos ;};
    bool operator != ( iter_function it)const{ return  Pos != it.Pos  ;};
    bool operator <  ( iter_function it)const{ return  Pos < it.Pos  ;};
    bool operator >  ( iter_function it)const{ return  Pos > it.Pos  ;};
    bool operator <= ( iter_function it)const{ return  Pos <= it.Pos  ;};
    bool operator >= ( iter_function it)const{ return  Pos >= it.Pos  ;};

    iter_function & operator ++ ( void) { ++Pos ; return *this ;};
    iter_function & operator -- ( void) { --Pos ; return *this ;};

    iter_function & operator += ( signed_type N){ Pos+= N ; return *this ;};
    iter_function & operator -= ( signed_type N){ Pos-= N ; return *this ;};

    iter_function operator + ( signed_type N )const
    {   return iter_function (Pos + N, function );
    };
    iter_function operator - ( signed_type N )const
    {   return iter_function (Pos - N, function );
    };
    signed_type operator - ( iter_function it) const { return (Pos- it.Pos);};

    value_t  operator * ( void){ return function();};
    const value_t operator * ( void)const { return function();};

    value_t operator -> ( void){ return function();};
    const value_t  operator -> ( void)const { return function();};
};

//##########################################################################
//                                                                        ##
//                 S T R U C T     L E S S _ P T R                        ##
//                                                                        ##
//##########################################################################
//---------------------------------------------------------------------------
/// @class less_ptr
///
/// @remarks this is the comparison object for pointers. Receive a object
///          for to compare the objects pointed. The pointers can be nullptr
//---------------------------------------------------------------------------
template <class T , class comp_t = std::less<T> >
struct less_ptr
{   //----------------------------- Variables -----------------------
    comp_t comp ;
    //----------------------------- Functions ----------------------
    inline less_ptr ( comp_t C1 = comp_t()):comp(C1){};
    inline bool operator ()( T * T1,  T * T2 ) const
    {   //-------------------- begin ------------------------------
        return  (T1==nullptr )?false:(T2==nullptr)?true:comp(*T1 ,*T2);
    };
};
//##########################################################################
//                                                                        ##
//         S T R U C T     L E S S _ P T R _ N O _ N U L L                ##
//                                                                        ##
//##########################################################################
//---------------------------------------------------------------------------
/// @class less_ptr_no_null
///
/// @remarks this is the comparison object for pointers. Receive a object
///          for to compare the objects pointed. The pointers can't be nullptr
//---------------------------------------------------------------------------
template    <   class iter_t ,
                class comp_t =std::less<typename iter_value<iter_t>::type>
            >
struct less_ptr_no_null
{   //----------------------------- Variables -----------------------
    comp_t comp ;
    //----------------------------- Functions ----------------------
    inline less_ptr_no_null ( comp_t C1 = comp_t()):comp(C1){};
    inline bool operator ()( iter_t  T1,  iter_t  T2 ) const
    {   //-------------------- begin ------------------------------
        return  comp(*T1 ,*T2);
    };
};
//****************************************************************************
}; //   Fin namespace util
}; //   Fin namespace countertree
}; //   Fin namespace boost
//****************************************************************************
#endif


